﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Problems.DataAnalysis.Symbolic;

namespace HeuristicLab.EvolutionaryTracking {
  public class SymbolicExpressionTreeGenealogyGraph : GenealogyGraph<ISymbolicExpressionTree> {
    public enum SimilarityLevel { Exact, High, Relaxed }

    private readonly Dictionary<int, Func<ISymbolicExpressionTreeNode, string>> _dispatcher;

    public delegate string HashFunction(ISymbolicExpressionTreeNode node);

    public SymbolicExpressionTreeGenealogyGraph() {
      _dispatcher = new Dictionary<int, Func<ISymbolicExpressionTreeNode, string>>();
      _dispatcher[(int)SimilarityLevel.Exact] = GetExactString;
      _dispatcher[(int)SimilarityLevel.High] = GetHighLevelString;
      _dispatcher[(int)SimilarityLevel.Relaxed] = GetRelaxedString;
    }

    #region Fragment tracing
    public IEnumerable<ISymbolicExpressionTree> TraceFragment(ISymbolicExpressionTreeNode fragment, int mode = 0) {
      return Keys.Where(tree => ContainsFragment(tree, fragment, mode) != -1);
    }

    public static bool CheckSimilarity(ISymbolicExpressionTree t1, ISymbolicExpressionTree t2, int mode) {
      return FindMatch(t1, t2, mode) == 0;
    }

    public static int ContainsFragment(ISymbolicExpressionTree tree, ISymbolicExpressionTreeNode fragment, int mode) {
      var nodes = tree.IterateNodesPostfix().ToArray();
      var fragments = fragment.IterateNodesPostfix().ToArray();
      return FindMatch(nodes, fragments, mode);
    }

    // Build a string hash out of all node symbols and parameter values (trimmed to 4 decimal places)
    private static string GetExactString(ISymbolicExpressionTreeNode node) {
      if (node is ConstantTreeNode) {
        var constant = node as ConstantTreeNode;
        return (constant.Symbol + string.Format("{0:0.0000}", constant.Value));
      }
      if (node is VariableTreeNode) {
        var variable = node as VariableTreeNode;
        return (variable.Symbol + variable.VariableName + variable.Weight);
      }
      return (node.Symbol.ToString());
    }

    // Build a string hash out of all node symbols in their respective order
    private static string GetHighLevelString(ISymbolicExpressionTreeNode node) {
      return (node.Symbol.ToString());
    }

    // Build a string hash which only considers function symbols and function arities. leaf nodes are taken as "wildcards"
    private static string GetRelaxedString(ISymbolicExpressionTreeNode node) {
      if (node.SubtreeCount != 0) /* function node */ {
        return (node.Symbol + node.SubtreeCount.ToString(CultureInfo.InvariantCulture));
      }
      return string.Empty;
    }

    #endregion
    public class SymbolicExpressionTreeNodeComparer : IEqualityComparer<ISymbolicExpressionTreeNode> {
      public SymbolicExpressionTreeNodeComparer(int similarityLevel) { _similarityLevel = similarityLevel; }

      public int GetHashCode(ISymbolicExpressionTreeNode n) { return n.ToString().ToLower().GetHashCode(); }

      public bool Equals(ISymbolicExpressionTreeNode a, ISymbolicExpressionTreeNode b) {
        if (a.SubtreeCount != b.SubtreeCount) { return false; }
        if (a.SubtreeCount > 0) { return a.Symbol.ToString().Equals(b.Symbol.ToString()); }
        switch (_similarityLevel) {
          case ((int)SimilarityLevel.Exact):
            if (a is ConstantTreeNode && b is ConstantTreeNode) { return (a as ConstantTreeNode).Value.Equals((b as ConstantTreeNode).Value); }
            if (a is VariableTreeNode && b is VariableTreeNode) { return (a as VariableTreeNode).Weight.Equals((b as VariableTreeNode).Weight); }
            return false;
          case ((int)SimilarityLevel.High):
            return (a is ConstantTreeNode && b is ConstantTreeNode) || (a is VariableTreeNode && b is VariableTreeNode);
          case ((int)SimilarityLevel.Relaxed):
            return true;
          default:
            return false;
        }
      }
      private readonly int _similarityLevel;
    }

    private static int FindMatch(ISymbolicExpressionTree t0, ISymbolicExpressionTree t1, int mode) {
      return FindMatch(t0.IterateNodesPostfix().ToArray(), t1.IterateNodesPostfix().ToArray(), mode);
    }

    private static int FindMatch(ISymbolicExpressionTreeNode t0, ISymbolicExpressionTreeNode t1, int mode) {
      return FindMatch(t0.IterateNodesPostfix().ToArray(), t1.IterateNodesPostfix().ToArray(), mode);
    }

    // returns the index in the sequence 'seq', where pattern 'pat' is matched, or -1 for no match
    // -1 means no match, 0 means perfect match (the sequences overlap, hence the match index is 0)
    public static int FindMatch(ISymbolicExpressionTreeNode[] seq, ISymbolicExpressionTreeNode[] pat, int mode) {
      int patlen = pat.Length;
      int seqlen = seq.Length;
      if (patlen == 0 || seqlen == 0) return -1;
      var comparer = new SymbolicExpressionTreeNodeComparer(mode);
      if (patlen == seqlen) return seq.SequenceEqual(pat, comparer) ? 0 : -1;
      int i = patlen;
      while (i <= seqlen) {
        var ch = seq[i - 1];
        if (comparer.Equals(ch, pat.Last())) {
          if (seq.Skip(i - patlen).Take(patlen).SequenceEqual(pat, comparer)) {
            return i - patlen;
          }
        }
        ++i;
      }
      return -1;
    }
  }
}
